/* Copyright (C) 1989, 1995, 1996 Aladdin Enterprises.  All rights reserved.
  
  This file is part of Aladdin Ghostscript.
  
  Aladdin Ghostscript is distributed with NO WARRANTY OF ANY KIND.  No author
  or distributor accepts any responsibility for the consequences of using it,
  or for whether it serves any particular purpose or works at all, unless he
  or she says so in writing.  Refer to the Aladdin Ghostscript Free Public
  License (the "License") for full details.
  
  Every copy of Aladdin Ghostscript must include a copy of the License,
  normally in a plain ASCII text file named PUBLIC.  The License grants you
  the right to copy, modify and redistribute Aladdin Ghostscript, but only
  under certain conditions described in the License.  Among other things, the
  License requires that the copyright notice and this notice be preserved on
  all copies.
*/

/* $Id: gdevepag.c,v 1.8 1998/11/25 05:24:22 nari Exp $ */
/* Many parts of this file are copied from gdevdjet.c */
/* Epson ESC/Page driver for Ghostscript */

#include "gdevprn.h"

#define ESC 0x1b    /* \033 */
#define GS  0x1d    /* \035 */

/* The device descriptors */
private dev_proc_open_device(epag_open);
private dev_proc_close_device(epag_close);
private dev_proc_print_page(epag_print_page);
private void epag_printer_initialize(gx_device_printer *pdev, FILE *fp,int);
private dev_proc_print_page(epag_print_page);
private int epag_flush(gx_device_printer *pdev, FILE *fp, int bpl, int cRow,
			byte *ibuf, byte *cbuf);
private dev_proc_get_params(epag_get_params);
private dev_proc_put_params(epag_put_params);
private void epag_paper_set(gx_device_printer *pdev, FILE *fp);

#define X_DPI 300
#define Y_DPI 300

static struct {
  bool tumble;
  bool noPaperSelect; /* ѻ極򥳡ɤϤʤ */
  float offX;   /* Xեåȡ ¦ ñ mm */
  float offY;   /* Yեåȡ ¦ ñ mm */
  int cRowBuf;
} epag_cont;

private gx_device_procs prn_epag_procs =
prn_params_procs(epag_open, gdev_prn_output_page, epag_close,
		 epag_get_params, epag_put_params);
gx_device_printer far_data gs_epag_device =
prn_device(prn_epag_procs, "epag", 
	   DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS,
	   X_DPI, Y_DPI,
	   0.0, 0.0, 0.0, 0.0,	/* margins */
	   1, epag_print_page);

#define ppdev ((gx_device_printer *)pdev)

/* Open the printer, adjusting the margins if necessary. */
private int
epag_open(gx_device *pdev)
{
  return gdev_prn_open(pdev);
}

private int
epag_close(gx_device *pdev)
{
  gdev_prn_open_printer(pdev, 1);
  if(ppdev->Duplex && (pdev->PageCount & 1)){
    fprintf(ppdev->file ,"%c%dY%c%dX", GS, 0, GS, 0);
    fprintf(ppdev->file,"Blank Page\r\n");
    fprintf(ppdev->file,"%c0dpsE",GS);
  }
  return gdev_prn_close(pdev);
}

/* #undef ppdev */

private int
epag_print_page(gx_device_printer *pdev, FILE *fp)
{
  int code = 0;
  int cRowBuf;            /* ٤˽Ϥ rowο */
  int cRow;               /* bufäƤrowο */
  int bpl;                /* byte per line */
  byte *ibuf,*ibuf2,*bp;  /* image buffer  */
  int x,y,ri;
  
  /* alloc buffers */

  if(epag_cont.cRowBuf < 1)
    epag_cont.cRowBuf = 128;
  cRowBuf = epag_cont.cRowBuf;
  
  bpl = gdev_mem_bytes_per_scan_line((gx_device *)pdev);
  ibuf = (byte *)gs_malloc(bpl , cRowBuf , "epag_print_page(ibuf)");
  if(ibuf == NULL) return_error(gs_error_VMerror);
  ibuf2 = (byte *)gs_malloc(bpl*3/2+1 , cRowBuf , "epag_print_page(ibuf2)");
  if(ibuf2 == NULL) return_error(gs_error_VMerror);

  /* printer initialize */
  
  if(pdev->PageCount == 0) 
    epag_printer_initialize(pdev, fp, 1);

  /* image output */
  
  bp = ibuf;
  cRow = 0;
  x = (epag_cont.offX / 25.4 - 0.2) * pdev->x_pixels_per_inch; 
  y = (epag_cont.offY / 25.4 - 0.2) * pdev->y_pixels_per_inch; 
  fprintf(fp ,"%c%dY%c%dX", GS, y, GS, x);
  for(ri = 0; ri < pdev->height; ri++){
    code = gdev_prn_copy_scan_lines(pdev, ri, bp, bpl);
    if(code < 0) return code;
    bp += bpl;
    if(++cRow >= cRowBuf){
      code = epag_flush(pdev, fp, bpl, cRow, ibuf, ibuf2);
      if(code < 0) return code;
      cRow = 0;
      bp = ibuf;
    }
  }
  if(cRow > 0){
    code = epag_flush(pdev, fp, bpl, cRow, ibuf, ibuf2);
    if(code < 0) return code;
  }

  if(pdev->Duplex)
    fprintf(fp,"%c0dpsE",GS);
  else
    fprintf(fp,"\014"); /* eject page */
  gs_free(ibuf2, bpl * 3 / 2 + 1, cRowBuf , "epag_print_page(ibuf2)");
  gs_free(ibuf, bpl, cRowBuf , "epag_print_page(ibuf)");
  return code;
}

/* output printer initialize code */

/* ------ Internal routines ------ */

static char can_inits[] ={
  ESC, 'S', ESC, ESC, 'S', 034,
  ESC, 'z', 0, 0,                  /* ESC Page ⡼ɤ  */
  GS, 'r', 'h', 'E',               /* hard reset         */
  GS, '1', 'm', 'm', 'E',          /* ڡ⡼ */
  GS, '3', 'b', 'c', 'I',          /* ̷           */
  GS, '2', 'a', 'b', 'P',          /* ᡼岼 */
};


private
void epag_printer_initialize(gx_device_printer *pdev, FILE *fp, int copies)
{
  double xDpi,yDpi;
  char *pse = NULL;
  int paper_size;
  
  xDpi = pdev->x_pixels_per_inch;
  yDpi = pdev->y_pixels_per_inch;

  /*  */
  
  fwrite(can_inits, sizeof(can_inits), 1, fp);
  
  /* ξ̰ */
  if(pdev->Duplex_set > 0){
    if(pdev->Duplex){
      fprintf(fp,"%c1sdE",GS);
      if(epag_cont.tumble==0)
	fprintf(fp,"%c0bdE",GS);
      else 
	fprintf(fp,"%c1bdE",GS);
    }
    else
      fprintf(fp,"%c0sdE",GS);
  }
    
  /* Ǿñ  */
  
  fprintf(fp,"%c0;%4.2fmuE",GS, 72.0 / xDpi );

  /*      */
  
  fprintf(fp,"%c0;%d;%ddrE",GS, (int)(xDpi+0.5),(int)(yDpi+0.5));

  /* ѻ極 */
  
  if(epag_cont.noPaperSelect == 0)
    epag_paper_set(pdev,fp);
  
  /* ԡ */

  if(copies > 1)
    fprintf(fp,"%c%dcoO",GS, copies < 256 ? copies : 255);
}

private
int epag_flush(gx_device_printer *pdev, FILE *fp, int bpl, int cRow,
		byte *ibuf, byte *cbuf)
{
  byte *cp,*bp,*bp1,*bp2,*bpMax;
  int  len;

  bp = ibuf;
  bpMax = ibuf + bpl * cRow;
  cp = cbuf;
  while(bp < bpMax){
    bp2 = bp + 256;
    if(bp2 > bpMax) bp2 = bpMax;
    for(bp1 = bp+1; (bp1 < bp2) && bp[0] == bp1[0]; bp1++);
    len = bp1 - bp;
    *cp++ = bp[0];
    if(len > 1) {
      *cp++ = bp[0];
      *cp++ = len - 2;
    }
    bp = bp1;
  }
  fprintf(fp,"%c%d;%d;%d;0bi{I",GS,cp-cbuf,bpl*8,cRow);
  fwrite(cbuf, cp-cbuf, 1, fp);
  return 0;
}

private int
epag_get_params(gx_device *pdev, gs_param_list *plist)
{
  int code;

  if(ppdev->Duplex_set < 0)
    ppdev->Duplex_set = 0;  /* ξ̰ǽˤ */

  if((code = gdev_prn_get_params(pdev, plist)) < 0 ||
     (code = param_write_int(plist, "cRowBuf", &epag_cont.cRowBuf)) < 0 ||
     (code = param_write_bool(plist, "Tumble", &epag_cont.tumble)) < 0 || 
     (code = param_write_bool(plist, "EpagNoPaperSelect",
			      &epag_cont.noPaperSelect)) < 0 || 
     (code = param_write_float(plist, "EpagOffX", &epag_cont.offX)) < 0 ||
     (code = param_write_float(plist, "EpagOffY", &epag_cont.offY)) < 0)
    return code;
  return code;
}

/* Put properties. */
private int
epag_put_params(gx_device *pdev, gs_param_list *plist)
{
  param_read_int(plist, "cRowBuf", &epag_cont.cRowBuf);
  param_read_bool(plist, "Tumble", &epag_cont.tumble);
  param_read_bool(plist, "EpagNoPaperSelect", &epag_cont.noPaperSelect);
  param_read_float(plist, "EpagOffX", &epag_cont.offX);
  param_read_float(plist, "EpagOffY", &epag_cont.offY);
  gdev_prn_put_params(pdev, plist);
  return 0;
}

typedef struct {
  int width;           /* ѻ  ñ point */
  int height;          /* ѻ  ñ point */
  int escpage;         /* ESC/Pageѻѥ᡼ */
} EpagPaperTable;

static EpagPaperTable epagPaperTable[] =
{
  { 842, 1190, 13 }, /* A3 */
  { 595,  842, 14 }, /* A4 */
  { 597,  842, 14 }, /* A4     8.3  11.7 */
  { 421,  595, 15 }, /* A5 */
  { 297,  421, 16 }, /* A6 */
  { 709, 1002, 24 }, /* B4 ISO */
  { 501,  709, 25 }, /* B5 ISO */
  { 729, 1032, 24 }, /* B4 JIS */
  { 516,  729, 25 }, /* B5 JIS */
  { 612,  792, 30 }, /* Letter */
  { 396,  612, 31 }, /* Half Letter */
  { 612, 1008, 32 }, /* Legal */
  { 522,  756, 33 }, /* Executive */
  { 612,  936, 34 }, /* Government Legal */
  { 576,  756, 35 }, /* Government Letter */
  { 792, 1224, 36 }, /* Ledger */
  { 593,  935, 37 }, /* F4 */
  { 284,  419, 38 }, /* Ϥ */
  { 933, 1369, 72 }, /* A3Υ */
  { 279,  540, 80 }, /* Monarch */
  { 297,  684, 81 }, /* Commercial 10 */
  { 312,  624, 90 }, /* DL */
  { 459,  649, 91 }, /* C5 */
  { 0,0,-1 }, /* 귿 */
};

void epag_paper_set(gx_device_printer *pdev, FILE *fp)
{
  int i,width,height,w,h,wp,hp,bLandscape;
  EpagPaperTable *pt;

  width  = pdev->MediaSize[0];
  height = pdev->MediaSize[1];

  if(width < height) {
    bLandscape = 0;
    w = width;
    h = height;
    wp = width  / 72.0 * pdev->x_pixels_per_inch;
    hp = height / 72.0 * pdev->y_pixels_per_inch;
  }
  else {
    bLandscape = 1;
    w = height;
    h = width;
    wp = height / 72.0 * pdev->y_pixels_per_inch;
    hp = width  / 72.0 * pdev->x_pixels_per_inch;
  }

  for(pt = epagPaperTable; pt->escpage > 0; pt++)
    if(pt->width == w && pt->height == h)
      break;

  if(pt->escpage > 0){
    fprintf(fp,"%c%d", GS, pt->escpage);
    if(pt->escpage < 0)
      fprintf(fp,";%d;%d",wp, hp);
    fprintf(fp,"psE");
    
    fprintf(fp,"%c%dpoE",GS,bLandscape);
  }
}

/*** Local Variables: ***/
/*** compile-command: "make -f unix-gcc.mak zlibc_=-lz PREFIX=/usr/local CFLAGS=\"-g -DA4\"" ***/
/*** End: ***/
